import { trace, context, SpanStatusCode, SpanKind } from '@opentelemetry/api';
import { getNodeAutoInstrumentations } from '@opentelemetry/auto-instrumentations-node';
import { OTLPTraceExporter } from '@opentelemetry/exporter-trace-otlp-http';
import { Resource } from '@opentelemetry/resources';
import { NodeSDK } from '@opentelemetry/sdk-node';
import { AlwaysOnSampler, AlwaysOffSampler, TraceIdRatioBasedSampler, ParentBasedSampler, ConsoleSpanExporter } from '@opentelemetry/sdk-trace-base';
import { ATTR_SERVICE_NAME } from '@opentelemetry/semantic-conventions';

// src/telemetry/telemetry.decorators.ts
function hasActiveTelemetry() {
  try {
    return !!trace.getTracer("default-tracer");
  } catch {
    return false;
  }
}

// src/telemetry/telemetry.decorators.ts
function withSpan(options) {
  return function(_target, propertyKey, descriptor) {
    if (!descriptor || typeof descriptor === "number") return;
    const originalMethod = descriptor.value;
    const methodName = String(propertyKey);
    descriptor.value = function(...args) {
      if (options?.skipIfNoTelemetry && !hasActiveTelemetry()) {
        return originalMethod.apply(this, args);
      }
      const tracer = trace.getTracer("default-tracer");
      let spanName;
      let spanKind;
      if (typeof options === "string") {
        spanName = options;
      } else if (options) {
        spanName = options.spanName || methodName;
        spanKind = options.spanKind;
      } else {
        spanName = methodName;
      }
      const span = tracer.startSpan(spanName, { kind: spanKind });
      const ctx = trace.setSpan(context.active(), span);
      args.forEach((arg, index) => {
        try {
          span.setAttribute(`${spanName}.argument.${index}`, JSON.stringify(arg));
        } catch (e) {
          span.setAttribute(`${spanName}.argument.${index}`, "[Not Serializable]");
        }
      });
      let result;
      try {
        result = context.with(ctx, () => originalMethod.apply(this, args));
        if (result instanceof Promise) {
          return result.then((resolvedValue) => {
            try {
              span.setAttribute(`${spanName}.result`, JSON.stringify(resolvedValue));
            } catch (e) {
              span.setAttribute(`${spanName}.result`, "[Not Serializable]");
            }
            return resolvedValue;
          }).finally(() => span.end());
        }
        try {
          span.setAttribute(`${spanName}.result`, JSON.stringify(result));
        } catch (e) {
          span.setAttribute(`${spanName}.result`, "[Not Serializable]");
        }
        return result;
      } catch (error) {
        span.setStatus({
          code: SpanStatusCode.ERROR,
          message: error instanceof Error ? error.message : "Unknown error"
        });
        if (error instanceof Error) {
          span.recordException(error);
        }
        throw error;
      } finally {
        if (!(result instanceof Promise)) {
          span.end();
        }
      }
    };
    return descriptor;
  };
}
function InstrumentClass(options) {
  return function(target) {
    const methods = Object.getOwnPropertyNames(target.prototype);
    methods.forEach((method) => {
      if (options?.excludeMethods?.includes(method) || method === "constructor") return;
      if (options?.methodFilter && !options.methodFilter(method)) return;
      const descriptor = Object.getOwnPropertyDescriptor(target.prototype, method);
      if (descriptor && typeof descriptor.value === "function") {
        Object.defineProperty(
          target.prototype,
          method,
          withSpan({
            spanName: options?.prefix ? `${options.prefix}.${method}` : method,
            skipIfNoTelemetry: true,
            spanKind: options?.spanKind || SpanKind.INTERNAL
          })(target, method, descriptor)
        );
      }
    });
    return target;
  };
}
var sdkInstance = null;
var Telemetry = class _Telemetry {
  sdk = null;
  tracer = trace.getTracer("default");
  name = "default-service";
  static isInitialized = false;
  getSampler(config) {
    if (!config.sampling) {
      return new AlwaysOnSampler();
    }
    if (!config.enabled) {
      return new AlwaysOffSampler();
    }
    switch (config.sampling.type) {
      case "ratio":
        return new TraceIdRatioBasedSampler(config.sampling.probability);
      case "always_on":
        return new AlwaysOnSampler();
      case "always_off":
        return new AlwaysOffSampler();
      case "parent_based":
        const rootSampler = new TraceIdRatioBasedSampler(config.sampling.root?.probability || 1);
        return new ParentBasedSampler({ root: rootSampler });
      default:
        return new AlwaysOnSampler();
    }
  }
  constructor(config) {
    this.name = config.serviceName ?? "default-service";
    if (typeof window === "undefined") {
      if (process.env.NODE_ENV === "development" || !sdkInstance) {
        if (sdkInstance) {
          this.shutdown();
        }
        const exporter = config.export?.type === "otlp" ? new OTLPTraceExporter({
          url: config.export.endpoint,
          headers: config.export.headers
        }) : config.export?.type === "custom" ? config.export.exporter : new ConsoleSpanExporter();
        const sampler = this.getSampler(config);
        sdkInstance = new NodeSDK({
          resource: new Resource({
            [ATTR_SERVICE_NAME]: this.name
          }),
          traceExporter: exporter,
          sampler,
          instrumentations: [getNodeAutoInstrumentations()]
        });
        try {
          sdkInstance.start();
          this.sdk = sdkInstance;
          _Telemetry.isInitialized = true;
        } catch (error) {
          console.warn("Failed to initialize OpenTelemetry:", error);
        }
      }
    }
    this.tracer = trace.getTracer(this.name);
  }
  async shutdown() {
    if (this.sdk && _Telemetry.isInitialized) {
      try {
        await this.sdk.shutdown();
        _Telemetry.isInitialized = false;
        global.__OTEL_SDK__ = void 0;
        global.__TELEMETRY__ = void 0;
      } catch (error) {
        console.warn("Error shutting down OpenTelemetry:", error);
      }
    }
  }
  /**
   * Initialize telemetry with the given configuration
   * @param config - Optional telemetry configuration object
   * @returns Telemetry instance that can be used for tracing
   */
  static init(config = {}) {
    try {
      if (!global.__TELEMETRY__) {
        global.__TELEMETRY__ = new _Telemetry(config);
      }
      return global.__TELEMETRY__;
    } catch (error) {
      console.error("Failed to initialize telemetry:", error);
      throw error;
    }
  }
  /**
   * Get the global telemetry instance
   * @throws {Error} If telemetry has not been initialized
   * @returns {Telemetry} The global telemetry instance
   */
  static get() {
    if (!global.__TELEMETRY__) {
      throw new Error("Telemetry not initialized");
    }
    return global.__TELEMETRY__;
  }
  /**
   * Wraps a class instance with telemetry tracing
   * @param instance The class instance to wrap
   * @param options Optional configuration for tracing
   * @returns Wrapped instance with all methods traced
   */
  traceClass(instance, options = {}) {
    const { skipIfNoTelemetry = true } = options;
    if (skipIfNoTelemetry && !hasActiveTelemetry()) {
      return instance;
    }
    const { spanNamePrefix = instance.constructor.name.toLowerCase(), attributes = {}, excludeMethods = [] } = options;
    return new Proxy(instance, {
      get: (target, prop) => {
        const value = target[prop];
        if (typeof value === "function" && prop !== "constructor" && !prop.toString().startsWith("_") && !excludeMethods.includes(prop.toString())) {
          return this.traceMethod(value.bind(target), {
            spanName: `${spanNamePrefix}.${prop.toString()}`,
            attributes: {
              ...attributes,
              [`${spanNamePrefix}.name`]: target.constructor.name,
              [`${spanNamePrefix}.method.name`]: prop.toString()
            }
          });
        }
        return value;
      }
    });
  }
  /**
   * method to trace individual methods with proper context
   * @param method The method to trace
   * @param context Additional context for the trace
   * @returns Wrapped method with tracing
   */
  traceMethod(method, context2) {
    const { skipIfNoTelemetry = true } = context2;
    if (skipIfNoTelemetry && !hasActiveTelemetry()) {
      return method;
    }
    return async (...args) => {
      const span = this.tracer.startSpan(context2.spanName);
      try {
        if (context2.attributes) {
          span.setAttributes(context2.attributes);
        }
        args.forEach((arg, index) => {
          try {
            span.setAttribute(`${context2.spanName}.argument.${index}`, JSON.stringify(arg));
          } catch (e) {
            span.setAttribute(`${context2.spanName}.argument.${index}`, "[Not Serializable]");
          }
        });
        const result = await method(...args);
        try {
          span.setAttribute(`${context2.spanName}.result`, JSON.stringify(result));
        } catch (e) {
          span.setAttribute(`${context2.spanName}.result`, "[Not Serializable]");
        }
        span.end();
        return result;
      } catch (error) {
        span.recordException(error);
        span.setStatus({
          code: SpanStatusCode.ERROR,
          message: error.message
        });
        span.end();
        throw error;
      }
    };
  }
};

export { InstrumentClass, Telemetry, hasActiveTelemetry, withSpan };
