import { MastraBase } from './chunk-TGPURYN5.js';
import { get } from 'radash';
import sift from 'sift';
import { setup, createActor, assign, fromPromise } from 'xstate';
import 'zod';

var Step = class {
  id;
  description;
  inputSchema;
  outputSchema;
  payload;
  execute;
  retryConfig;
  mastra;
  constructor({
    id,
    description,
    execute,
    payload,
    outputSchema,
    inputSchema,
    retryConfig
  }) {
    this.id = id;
    this.description = description ?? "";
    this.inputSchema = inputSchema;
    this.payload = payload;
    this.outputSchema = outputSchema;
    this.execute = execute;
    this.retryConfig = retryConfig;
  }
};
function createStep(opts) {
  return new Step(opts);
}

// src/workflows/utils.ts
function isErrorEvent(stateEvent) {
  return stateEvent.type.startsWith("xstate.error.actor.");
}
function isTransitionEvent(stateEvent) {
  return stateEvent.type.startsWith("xstate.done.actor.");
}
function isVariableReference(value) {
  return typeof value === "object" && "step" in value && "path" in value;
}
function getStepResult(result) {
  if (result?.status === "success") return result.output;
  return undefined;
}

// src/workflows/workflow.ts
var Workflow = class extends MastraBase {
  name;
  triggerSchema;
  /** XState machine instance that orchestrates the workflow execution */
  #machine;
  /** XState actor instance that manages the workflow execution */
  #actor = null;
  #runId;
  #retryConfig;
  #mastra;
  // registers stepIds on `after` calls
  #afterStepStack = [];
  #lastStepStack = [];
  #stepGraph = { initial: [] };
  #stepSubscriberGraph = {};
  #steps = {};
  /**
   * Creates a new Workflow instance
   * @param name - Identifier for the workflow (not necessarily unique)
   * @param logger - Optional logger instance
   */
  constructor({ name, triggerSchema, retryConfig, mastra }) {
    super({ component: "WORKFLOW", name });
    this.name = name;
    this.#retryConfig = retryConfig;
    this.triggerSchema = triggerSchema;
    this.#runId = crypto.randomUUID();
    this.#mastra = mastra;
    if (mastra?.logger) {
      this.logger = mastra?.logger;
    }
    this.initializeMachine();
  }
  /**
   * Initializes the XState machine for the workflow
   *
   * Registers the machine's types, actions, actors, initial context, entry actions, initial state, and states
   * @returns The initialized machine
   */
  initializeMachine() {
    const machine = setup({
      types: {},
      delays: this.#makeDelayMap(),
      actions: this.#getDefaultActions(),
      actors: this.#getDefaultActors()
    }).createMachine({
      id: this.name,
      type: "parallel",
      context: ({ input }) => ({
        ...input
      }),
      states: this.#buildStateHierarchy(this.#stepGraph)
    });
    this.#machine = machine;
    return machine;
  }
  step(step, config) {
    const { variables = {} } = config || {};
    const requiredData = {};
    for (const [key, variable] of Object.entries(variables)) {
      if (variable && isVariableReference(variable)) {
        requiredData[key] = variable;
      }
    }
    const stepKey = this.#makeStepKey(step);
    const graphEntry = {
      step,
      config: {
        ...this.#makeStepDef(stepKey),
        ...config,
        data: requiredData
      }
    };
    this.#steps[stepKey] = step;
    const parentStepKey = this.#afterStepStack[this.#afterStepStack.length - 1];
    const stepGraph = this.#stepSubscriberGraph[parentStepKey || ""];
    if (parentStepKey && stepGraph) {
      if (!stepGraph.initial.some((step2) => step2.step.id === stepKey)) {
        stepGraph.initial.push(graphEntry);
      }
      stepGraph[stepKey] = [];
    } else {
      if (!this.#stepGraph[stepKey]) this.#stepGraph[stepKey] = [];
      this.#stepGraph.initial.push(graphEntry);
    }
    this.#lastStepStack.push(stepKey);
    return this;
  }
  then(step, config) {
    const { variables = {} } = config || {};
    const requiredData = {};
    for (const [key, variable] of Object.entries(variables)) {
      if (variable && isVariableReference(variable)) {
        requiredData[key] = variable;
      }
    }
    const lastStepKey = this.#lastStepStack[this.#lastStepStack.length - 1];
    const stepKey = this.#makeStepKey(step);
    const graphEntry = {
      step,
      config: {
        ...this.#makeStepDef(stepKey),
        ...config,
        data: requiredData
      }
    };
    this.#steps[stepKey] = step;
    if (!lastStepKey) return this;
    const parentStepKey = this.#afterStepStack[this.#afterStepStack.length - 1];
    const stepGraph = this.#stepSubscriberGraph[parentStepKey || ""];
    if (parentStepKey && stepGraph && stepGraph[lastStepKey]) {
      stepGraph[lastStepKey].push(graphEntry);
    } else {
      if (!this.#stepGraph[lastStepKey]) this.#stepGraph[lastStepKey] = [];
      this.#stepGraph[lastStepKey].push(graphEntry);
    }
    return this;
  }
  after(step) {
    const stepKey = this.#makeStepKey(step);
    this.#afterStepStack.push(stepKey);
    if (!this.#stepSubscriberGraph[stepKey]) {
      this.#stepSubscriberGraph[stepKey] = { initial: [] };
    }
    return this;
  }
  /**
   * Executes the workflow with the given trigger data
   * @param triggerData - Initial data to start the workflow with
   * @returns Promise resolving to workflow results or rejecting with error
   * @throws Error if trigger schema validation fails
   */
  createRun() {
    const runId = crypto.randomUUID();
    this.#runId = runId;
    return {
      runId,
      start: async ({ triggerData } = {}) => this.execute({ triggerData })
    };
  }
  async execute({
    triggerData,
    snapshot,
    runId,
    stepId
  } = {}) {
    if (runId) {
      this.#runId = runId;
      this.logger.debug("Incoming snapshot", {
        snapshot,
        runId: this.#runId
      });
    }
    const machineInput = snapshot ? snapshot.context : {
      // Maintain the original step results and their output
      steps: {},
      triggerData: triggerData || {},
      attempts: Object.keys(this.#steps).reduce(
        (acc, stepKey) => {
          acc[stepKey] = this.#steps[stepKey]?.retryConfig?.attempts || this.#retryConfig?.attempts || 3;
          return acc;
        },
        {}
      )
    };
    this.logger.debug("Machine input prepared", {
      machineInput,
      runId: this.#runId
    });
    const actorSnapshot = snapshot ? {
      ...snapshot,
      context: machineInput
    } : undefined;
    this.logger.debug("Creating actor with configuration", {
      machineInput,
      actorSnapshot,
      machineStates: this.#machine.config.states,
      runId: this.#runId
    });
    this.#actor = createActor(this.#machine, {
      inspect: (inspectionEvent) => {
        this.logger.debug("XState inspection event", {
          type: inspectionEvent.type,
          event: inspectionEvent.event,
          runId: this.#runId
        });
      },
      input: machineInput,
      snapshot: actorSnapshot
    });
    this.#actor.start();
    if (stepId) {
      this.#actor.send({ type: "RESET_TO_PENDING", stepId });
    }
    this.logger.debug("Actor started", { runId: this.#runId });
    return new Promise((resolve, reject) => {
      if (!this.#actor) {
        reject(new Error("Actor not initialized"));
        return;
      }
      const suspendedPaths = /* @__PURE__ */ new Set();
      this.#actor.subscribe(async (state) => {
        this.#getSuspendedPaths({
          value: state.value,
          path: "",
          suspendedPaths
        });
        const allStatesValue = state.value;
        const allStatesComplete = this.#recursivelyCheckForFinalState({
          value: allStatesValue,
          suspendedPaths,
          path: ""
        });
        this.logger.debug("State completion check", {
          allStatesComplete,
          suspendedPaths: Array.from(suspendedPaths),
          runId: this.#runId
        });
        if (!allStatesComplete) return;
        try {
          await this.#persistWorkflowSnapshot();
          this.#cleanup();
          resolve({
            triggerData,
            results: state.context.steps,
            runId: this.#runId
          });
        } catch (error) {
          this.logger.debug("Failed to persist final snapshot", { error });
          this.#cleanup();
          resolve({
            triggerData,
            results: state.context.steps,
            runId: this.#runId
          });
        }
      });
    });
  }
  /**
   * Rebuilds the machine with the current steps configuration and validates the workflow
   *
   * This is the last step of a workflow builder method chain
   * @throws Error if validation fails
   *
   * @returns this instance for method chaining
   */
  commit() {
    this.initializeMachine();
    return this;
  }
  // record all object paths that leads to a suspended state
  #getSuspendedPaths({
    value,
    path,
    suspendedPaths
  }) {
    if (typeof value === "string") {
      if (value === "suspended") {
        suspendedPaths.add(path);
      }
    } else {
      Object.keys(value).forEach(
        (key) => this.#getSuspendedPaths({ value: value[key], path: path ? `${path}.${key}` : key, suspendedPaths })
      );
    }
  }
  #isFinalState(status) {
    return ["completed", "failed"].includes(status);
  }
  #recursivelyCheckForFinalState({
    value,
    suspendedPaths,
    path
  }) {
    if (typeof value === "string") {
      return this.#isFinalState(value) || suspendedPaths.has(path);
    }
    return Object.keys(value).every(
      (key) => this.#recursivelyCheckForFinalState({ value: value[key], suspendedPaths, path: path ? `${path}.${key}` : key })
    );
  }
  #buildBaseState(stepNode, nextSteps = []) {
    const nextStep = nextSteps.shift();
    return {
      initial: "pending",
      on: {
        RESET_TO_PENDING: {
          target: ".pending"
          // Note the dot to target child state
        }
      },
      states: {
        pending: {
          entry: () => {
            this.logger.debug(`Step ${stepNode.step.id} pending`);
          },
          exit: () => {
            this.logger.debug(`Step ${stepNode.step.id} finished pending`);
          },
          invoke: {
            src: "conditionCheck",
            input: ({ context }) => {
              return {
                context,
                stepNode
              };
            },
            onDone: [
              {
                guard: ({ event }) => {
                  return event.output.type === "SUSPENDED";
                },
                target: "suspended",
                actions: [
                  assign({
                    steps: ({ context, event }) => {
                      if (event.output.type !== "SUSPENDED") return context.steps;
                      return {
                        ...context.steps,
                        [stepNode.step.id]: {
                          status: "suspended",
                          ...context.steps?.[stepNode.step.id] || {}
                        }
                      };
                    },
                    attempts: ({ context, event }) => {
                      if (event.output.type !== "SUSPENDED") return context.attempts;
                      return { ...context.attempts, [stepNode.step.id]: stepNode.step.retryConfig?.attempts || 3 };
                    }
                  })
                ]
              },
              {
                guard: ({ event }) => {
                  return event.output.type === "WAITING";
                },
                target: "waiting",
                actions: [
                  { type: "decrementAttemptCount", params: { stepId: stepNode.step.id } },
                  assign({
                    steps: ({ context, event }) => {
                      if (event.output.type !== "WAITING") return context.steps;
                      return {
                        ...context.steps,
                        [stepNode.step.id]: {
                          status: "waiting"
                        }
                      };
                    }
                  })
                ]
              },
              {
                guard: ({ event }) => {
                  return event.output.type === "CONDITIONS_MET";
                },
                target: "executing"
              },
              {
                guard: ({ event }) => {
                  return event.output.type === "CONDITION_FAILED";
                },
                target: "failed",
                actions: assign({
                  steps: ({ context, event }) => {
                    if (event.output.type !== "CONDITION_FAILED") return context.steps;
                    this.logger.debug(`workflow condition check failed`, {
                      error: event.output.error,
                      stepId: stepNode.step.id
                    });
                    return {
                      ...context.steps,
                      [stepNode.step.id]: {
                        status: "failed",
                        error: event.output.error
                      }
                    };
                  }
                })
              }
            ]
          }
        },
        waiting: {
          entry: () => {
            this.logger.debug(`Step ${stepNode.step.id} waiting ${(/* @__PURE__ */ new Date()).toISOString()}`);
          },
          exit: () => {
            this.logger.debug(`Step ${stepNode.step.id} finished waiting ${(/* @__PURE__ */ new Date()).toISOString()}`);
          },
          after: {
            [stepNode.step.id]: {
              target: "pending"
            }
          }
        },
        suspended: {
          type: "final",
          entry: [
            () => {
              this.logger.debug(`Step ${stepNode.step.id} suspended`);
            },
            assign({
              steps: ({ context }) => ({
                ...context.steps,
                [stepNode.step.id]: {
                  ...context?.steps?.[stepNode.step.id] || {},
                  status: "suspended"
                }
              })
            })
          ]
          // after: {
          //   [stepNode.step.id]: {
          //     target: 'pending',
          //     actions: [
          //       assign({
          //         attempts: ({ context }: { context: WorkflowContext }) => ({
          //           ...context.attempts,
          //           [stepNode.step.id]: this.#steps[stepNode.step.id]?.retryConfig?.attempts || 3,
          //         }),
          //       }),
          //     ],
          //   }
          // },
          // entry: () => {
          //   this.logger.debug(`Step ${stepNode.step.id} suspended ${new Date().toISOString()}`);
          // },
          // exit: () => {
          //   this.logger.debug(`Step ${stepNode.step.id} finished suspended ${new Date().toISOString()}`);
          // },
          // after: {
          //   [stepNode.step.id]: {
          //     target: 'suspended',
          //   },
          // },
        },
        executing: {
          entry: () => {
            this.logger.debug(`Step ${stepNode.step.id} executing`);
          },
          on: {
            SUSPENDED: {
              target: "suspended",
              actions: [
                assign({
                  steps: ({ context }) => ({
                    ...context.steps,
                    [stepNode.step.id]: {
                      status: "suspended"
                    }
                  })
                })
              ]
            }
          },
          invoke: {
            src: "resolverFunction",
            input: ({ context }) => ({
              context,
              stepNode
            }),
            onDone: {
              target: "runningSubscribers",
              actions: [
                ({ event }) => this.logger.debug(`Step ${stepNode.step.id} finished executing`, { output: event.output }),
                { type: "updateStepResult", params: { stepId: stepNode.step.id } },
                { type: "spawnSubscribers", params: { stepId: stepNode.step.id } }
              ]
            },
            onError: {
              target: "failed",
              actions: [{ type: "setStepError", params: { stepId: stepNode.step.id } }]
            }
          }
        },
        runningSubscribers: {
          entry: () => {
            this.logger.debug(`Step ${stepNode.step.id} running subscribers`);
          },
          exit: () => {
            this.logger.debug(`Step ${stepNode.step.id} finished running subscribers`);
          },
          invoke: {
            src: "spawnSubscriberFunction",
            input: ({ context }) => ({
              parentStepId: stepNode.step.id,
              context
            }),
            onDone: {
              target: nextStep ? nextStep.step.id : "completed",
              actions: [
                assign({
                  steps: ({ context, event }) => ({
                    ...context.steps,
                    ...event.output.steps
                  })
                }),
                () => this.logger.debug(`Subscriber execution completed`, { stepId: stepNode.step.id })
              ]
            },
            onError: {
              target: nextStep ? nextStep.step.id : "completed",
              actions: ({ event }) => {
                this.logger.debug(`Subscriber execution failed`, {
                  error: event.error,
                  stepId: stepNode.step.id
                });
              }
            }
          }
        },
        completed: {
          type: "final",
          entry: [
            { type: "notifyStepCompletion", params: { stepId: stepNode.step.id } },
            { type: "snapshotStep", params: { stepId: stepNode.step.id } },
            { type: "persistSnapshot" }
          ]
        },
        failed: {
          type: "final",
          entry: [
            { type: "notifyStepCompletion", params: { stepId: stepNode.step.id } },
            { type: "snapshotStep", params: { stepId: stepNode.step.id } },
            { type: "persistSnapshot" }
          ]
        },
        // build chain of next steps recursively
        ...nextStep ? { [nextStep.step.id]: { ...this.#buildBaseState(nextStep, nextSteps) } } : {}
      }
    };
  }
  #makeStepKey(step) {
    return `${step.id}`;
  }
  /**
   * Builds the state hierarchy for the workflow
   * @returns Object representing the state hierarchy
   */
  #buildStateHierarchy(stepGraph) {
    const states = {};
    stepGraph.initial.forEach((stepNode) => {
      const nextSteps = [...stepGraph[stepNode.step.id] || []];
      states[stepNode.step.id] = {
        ...this.#buildBaseState(stepNode, nextSteps)
      };
    });
    return states;
  }
  #getDefaultActions() {
    return {
      updateStepResult: assign({
        steps: ({ context, event }) => {
          if (!isTransitionEvent(event)) return context.steps;
          const { stepId, result } = event.output;
          return {
            ...context.steps,
            [stepId]: {
              status: "success",
              output: result
            }
          };
        }
      }),
      setStepError: assign({
        steps: ({ context, event }, params) => {
          if (!isErrorEvent(event)) return context.steps;
          const { stepId } = params;
          if (!stepId) return context.steps;
          return {
            ...context.steps,
            [stepId]: {
              status: "failed",
              error: event.error.message
            }
          };
        }
      }),
      notifyStepCompletion: async (_, params) => {
        const { stepId } = params;
        this.logger.debug(`Step ${stepId} completed`);
      },
      snapshotStep: assign({
        _snapshot: ({}, params) => {
          const { stepId } = params;
          return { stepId };
        }
      }),
      persistSnapshot: async ({ context }) => {
        if (context._snapshot) {
          return await this.#persistWorkflowSnapshot();
        }
        return;
      },
      decrementAttemptCount: assign({
        attempts: ({ context, event }, params) => {
          if (!isTransitionEvent(event)) return context.attempts;
          const { stepId } = params;
          const attemptCount = context.attempts[stepId];
          if (attemptCount === undefined) return context.attempts;
          return { ...context.attempts, [stepId]: attemptCount - 1 };
        }
      })
    };
  }
  #getInjectables() {
    return {
      runId: this.#runId,
      mastra: this.#mastra
    };
  }
  #getDefaultActors() {
    return {
      resolverFunction: fromPromise(async ({ input }) => {
        const { stepNode, context } = input;
        const injectables = this.#getInjectables();
        const resolvedData = this.#resolveVariables({
          stepConfig: stepNode.config,
          context,
          stepId: stepNode.step.id
        });
        this.logger.debug(`Resolved variables for ${stepNode.step.id}`, {
          resolvedData,
          runId: this.#runId
        });
        const result = await stepNode.config.handler({
          context: resolvedData,
          suspend: async () => {
            if (this.#actor) {
              context.steps[stepNode.step.id] = {
                status: "suspended"
              };
              await this.#persistWorkflowSnapshot();
              this.logger.debug(`Sending SUSPENDED event for step ${stepNode.step.id}`);
              this.#actor?.send({ type: "SUSPENDED", stepId: stepNode.step.id });
            } else {
              this.logger.debug(`Actor not available for step ${stepNode.step.id}`);
            }
          },
          ...injectables
        });
        this.logger.debug(`Step ${stepNode.step.id} result`, {
          result,
          runId: this.#runId
        });
        return {
          stepId: stepNode.step.id,
          result
        };
      }),
      conditionCheck: fromPromise(async ({ input }) => {
        const { context, stepNode } = input;
        const stepConfig = stepNode.config;
        const attemptCount = context.attempts[stepNode.step.id];
        this.logger.debug(`Checking conditions for ${stepNode.step.id}`, {
          stepId: stepNode.step.id,
          runId: this.#runId
        });
        this.logger.debug(`Attempt count for ${stepNode.step.id}`, {
          attemptCount,
          attempts: context.attempts,
          runId: this.#runId,
          stepId: stepNode.step.id
        });
        if (!attemptCount || attemptCount < 0) {
          if (stepConfig?.snapshotOnTimeout) {
            return { type: "SUSPENDED", stepId: stepNode.step.id };
          }
          return { type: "CONDITION_FAILED", error: `Step:${stepNode.step.id} condition check failed` };
        }
        if (!stepConfig?.when) {
          return { type: "CONDITIONS_MET" };
        }
        this.logger.debug(`Checking conditions for ${stepNode.step.id}`);
        if (typeof stepConfig?.when === "function") {
          const conditionMet = await stepConfig.when({ context });
          if (conditionMet) {
            this.logger.debug(`Condition met for ${stepNode.step.id}`);
            return { type: "CONDITIONS_MET" };
          }
          if (!attemptCount || attemptCount < 0) {
            return { type: "CONDITION_FAILED", error: `Step:${stepNode.step.id} condition check failed` };
          }
          return { type: "WAITING", stepId: stepNode.step.id };
        } else {
          const conditionMet = this.#evaluateCondition(stepConfig.when, context);
          if (!conditionMet) {
            return {
              type: "CONDITION_FAILED",
              error: `Step:${stepNode.step.id} condition check failed`
            };
          }
        }
        return { type: "CONDITIONS_MET" };
      }),
      spawnSubscriberFunction: fromPromise(
        async ({
          input
        }) => {
          const { parentStepId, context } = input;
          const stepGraph = this.#stepSubscriberGraph[parentStepId];
          if (!stepGraph) {
            return {
              steps: {}
            };
          }
          const subscriberMachine = setup({
            types: {},
            delays: this.#makeDelayMap(),
            actions: this.#getDefaultActions(),
            actors: this.#getDefaultActors()
          }).createMachine({
            id: `${this.name}-subscriber-${parentStepId}`,
            context,
            type: "parallel",
            states: this.#buildStateHierarchy(stepGraph)
          });
          const actor = createActor(subscriberMachine, { input: context });
          actor.start();
          return new Promise((resolve) => {
            const suspendedPaths = /* @__PURE__ */ new Set();
            actor.subscribe((state) => {
              this.#getSuspendedPaths({
                value: state.value,
                path: "",
                suspendedPaths
              });
              const allStatesValue = state.value;
              const allStatesComplete = this.#recursivelyCheckForFinalState({
                value: allStatesValue,
                suspendedPaths,
                path: ""
              });
              if (allStatesComplete) {
                actor.stop();
                resolve({
                  steps: state.context.steps
                });
              }
            });
          });
        }
      )
    };
  }
  /**
   * Persists the workflow state to the database
   */
  async #persistWorkflowSnapshot() {
    const snapshotFromActor = this.#actor?.getPersistedSnapshot();
    if (!this.#mastra?.storage) {
      this.logger.debug("Snapshot cannot be persisted. Mastra engine is not initialized", {
        runId: this.#runId
      });
      return;
    }
    if (!snapshotFromActor) {
      this.logger.debug("Snapshot cannot be persisted. No snapshot received.", { runId: this.#runId });
      return;
    }
    if (this.#mastra?.storage) {
      await this.#mastra.storage.persistWorkflowSnapshot({
        workflowName: this.name,
        runId: this.#runId,
        snapshot: snapshotFromActor
      });
    }
    return this.#runId;
  }
  async #loadWorkflowSnapshot(runId) {
    if (!this.#mastra?.storage) {
      this.logger.debug("Snapshot cannot be loaded. Mastra engine is not initialized", { runId });
      return;
    }
    return this.#mastra.storage.loadWorkflowSnapshot({ runId, workflowName: this.name });
  }
  /**
   * Resolves variables for a step from trigger data or previous step results
   * @param stepConfig - Configuration of the step needing variable resolution
   * @param context - Current workflow context containing results and trigger data
   * @returns Object containing resolved variable values
   */
  #resolveVariables({
    stepConfig,
    context,
    stepId
  }) {
    this.logger.debug(`Resolving variables for ${stepId}`, {
      runId: this.#runId
    });
    const resolvedData = {
      ...context,
      getStepPayload: (stepId2) => {
        if (stepId2 === "trigger") {
          return context.triggerData;
        }
        const result = context.steps[stepId2];
        if (result && result.status === "success") {
          return result.output;
        }
        return undefined;
      }
    };
    for (const [key, variable] of Object.entries(stepConfig.data)) {
      const sourceData = variable.step === "trigger" ? context.triggerData : getStepResult(context.steps[variable.step.id]);
      this.logger.debug(
        `Got source data for ${key} variable from ${variable.step === "trigger" ? "trigger" : variable.step.id}`,
        {
          sourceData,
          path: variable.path,
          runId: this.#runId
        }
      );
      if (!sourceData && variable.step !== "trigger") {
        resolvedData[key] = undefined;
        continue;
      }
      const value = variable.path === "" || variable.path === "." ? sourceData : get(sourceData, variable.path);
      this.logger.debug(`Resolved variable ${key}`, {
        value,
        runId: this.#runId
      });
      resolvedData[key] = value;
    }
    return resolvedData;
  }
  /**
   * Evaluates a single condition against workflow context
   */
  #evaluateCondition(condition, context) {
    let andBranchResult = true;
    let baseResult = true;
    let orBranchResult = true;
    const simpleCondition = Object.entries(condition).find(([key]) => key.includes("."));
    if (simpleCondition) {
      const [key, queryValue] = simpleCondition;
      const [stepId, ...pathParts] = key.split(".");
      const path = pathParts.join(".");
      const sourceData = stepId === "trigger" ? context.triggerData : getStepResult(context.steps[stepId]);
      this.logger.debug(`Got condition data from ${stepId}`, {
        sourceData,
        runId: this.#runId
      });
      if (!sourceData) {
        return false;
      }
      let value = get(sourceData, path);
      if (stepId !== "trigger" && path === "status" && !value) {
        value = "success";
      }
      if (typeof queryValue === "object" && queryValue !== null) {
        baseResult = sift(queryValue)(value);
      } else {
        baseResult = value === queryValue;
      }
    }
    if ("ref" in condition) {
      const { ref, query } = condition;
      const sourceData = ref.step === "trigger" ? context.triggerData : getStepResult(context.steps[ref.step.id]);
      this.logger.debug(`Got condition data from ${ref.step === "trigger" ? "trigger" : ref.step.id}`, {
        sourceData,
        runId: this.#runId
      });
      if (!sourceData) {
        return false;
      }
      let value = get(sourceData, ref.path);
      if (ref.step !== "trigger" && ref.path === "status" && !value) {
        value = "success";
      }
      baseResult = sift(query)(value);
    }
    if ("and" in condition) {
      andBranchResult = condition.and.every((cond) => this.#evaluateCondition(cond, context));
      this.logger.debug(`Evaluated AND condition`, {
        andBranchResult,
        runId: this.#runId
      });
    }
    if ("or" in condition) {
      orBranchResult = condition.or.some((cond) => this.#evaluateCondition(cond, context));
      this.logger.debug(`Evaluated OR condition`, {
        orBranchResult,
        runId: this.#runId
      });
    }
    const finalResult = baseResult && andBranchResult && orBranchResult;
    this.logger.debug(`Evaluated condition`, {
      finalResult,
      runId: this.#runId
    });
    return finalResult;
  }
  #makeStepDef(stepId) {
    const telemetry = this.#mastra?.telemetry;
    const handler = async ({ context, ...rest }) => {
      const targetStep = this.#steps[stepId];
      if (!targetStep) throw new Error(`Step not found`);
      const { payload = {}, execute } = targetStep;
      const mergedData = {
        ...payload,
        ...context
      };
      const finalAction = telemetry ? telemetry.traceMethod(execute, {
        spanName: `workflow.${this.name}.action.${stepId}`
      }) : execute;
      return finalAction ? await finalAction({ context: mergedData, ...rest }) : {};
    };
    const finalHandler = telemetry ? telemetry.traceMethod(handler, {
      spanName: `workflow.${this.name}.step.${stepId}`
    }) : handler;
    return {
      handler: finalHandler,
      data: {}
    };
  }
  /**
   * Creates a map of step IDs to their respective delay values
   * @returns Object mapping step IDs to delay values
   */
  #makeDelayMap() {
    const delayMap = {};
    Object.keys(this.#steps).forEach((stepId) => {
      delayMap[stepId] = this.#steps[stepId]?.retryConfig?.delay || this.#retryConfig?.delay || 1e3;
    });
    return delayMap;
  }
  /**
   * Cleans up the actor instance
   */
  #cleanup() {
    if (this.#actor) {
      this.#actor.stop();
      this.#actor = null;
    }
  }
  #getActivePathsAndStatus(value) {
    const paths = [];
    const traverse = (current, path = []) => {
      for (const [key, value2] of Object.entries(current)) {
        const currentPath = [...path, key];
        if (typeof value2 === "string") {
          paths.push({
            stepPath: currentPath,
            stepId: key,
            status: value2
          });
        } else if (typeof value2 === "object" && value2 !== null) {
          traverse(value2, currentPath);
        }
      }
    };
    traverse(value);
    return paths;
  }
  async getState(runId) {
    if (this.#runId === runId && this.#actor) {
      const snapshot = this.#actor.getSnapshot();
      const m = this.#getActivePathsAndStatus(snapshot.value);
      return {
        runId,
        value: snapshot.value,
        context: snapshot.context,
        activePaths: m,
        timestamp: Date.now()
      };
    }
    const storedSnapshot = await this.#mastra?.storage?.loadWorkflowSnapshot({
      runId,
      workflowName: this.name
    });
    if (storedSnapshot) {
      const parsed = storedSnapshot;
      const m = this.#getActivePathsAndStatus(parsed.value);
      return {
        runId,
        value: parsed.value,
        context: parsed.context,
        activePaths: m,
        timestamp: Date.now()
      };
    }
    return null;
  }
  #hasStateChanged(previous, current) {
    if (!previous) return true;
    const previousPaths = previous.activePaths;
    const currentPaths = current.activePaths;
    if (previousPaths.length !== currentPaths.length) return true;
    for (let i = 0; i < currentPaths.length; i++) {
      const prevPath = previousPaths[i];
      const currPath = currentPaths[i];
      if (JSON.stringify(prevPath?.stepPath) !== JSON.stringify(currPath?.stepPath)) return true;
      if (prevPath?.status !== currPath?.status) return true;
    }
    if (JSON.stringify(previous.context.steps) !== JSON.stringify(current.context.steps)) return true;
    return false;
  }
  #getDeepestState(value) {
    let deepestBranch = null;
    const traverse = (obj, path = []) => {
      for (const [key, value2] of Object.entries(obj)) {
        if (typeof value2 === "string") {
          const depth = path.length;
          if (!deepestBranch || depth > deepestBranch.depth) {
            deepestBranch = { stepId: key, status: value2, depth };
          }
        } else if (typeof value2 === "object" && value2 !== null) {
          traverse(value2, [...path, key]);
        }
      }
    };
    traverse(value);
    if (!deepestBranch) {
      throw new Error("No valid state found");
    }
    return {
      stepId: deepestBranch?.stepId,
      status: deepestBranch?.status
    };
  }
  async watch(runId, options = {}) {
    const { onTransition, pollInterval = 1e3 } = options;
    return new Promise((resolve, reject) => {
      let timeoutId;
      let previousState = null;
      const poll = async () => {
        try {
          const currentState = await this.getState(runId);
          if (!currentState) {
            reject(new Error(`No workflow found for runId: ${runId}`));
            return;
          }
          if (onTransition && this.#hasStateChanged(previousState, currentState)) {
            await onTransition(currentState);
          }
          previousState = currentState;
          const deepState = this.#getDeepestState(currentState.value);
          if (this.#isFinalState(deepState.status)) {
            resolve(currentState);
            return;
          }
          timeoutId = setTimeout(poll, pollInterval);
        } catch (error) {
          reject(error);
        }
      };
      poll();
      return () => {
        if (timeoutId) {
          clearTimeout(timeoutId);
        }
      };
    });
  }
  async resume({
    runId,
    stepId,
    context: resumeContext
  }) {
    const snapshot = await this.#loadWorkflowSnapshot(runId);
    if (!snapshot) {
      throw new Error(`No snapshot found for workflow run ${runId}`);
    }
    let parsedSnapshot;
    try {
      parsedSnapshot = typeof snapshot === "string" ? JSON.parse(snapshot) : snapshot;
    } catch (error) {
      this.logger.debug("Failed to parse workflow snapshot for resume", { error, runId });
      throw new Error("Failed to parse workflow snapshot");
    }
    if (resumeContext) {
      parsedSnapshot.context.steps[stepId] = {
        status: "success",
        output: {
          ...parsedSnapshot?.context?.steps?.[stepId]?.output || {},
          ...resumeContext
        }
      };
    }
    if (parsedSnapshot.children) {
      Object.entries(parsedSnapshot.children).forEach(([_childId, child]) => {
        if (child.snapshot?.input?.stepNode) {
          const stepDef = this.#makeStepDef(child.snapshot.input.stepNode.step.id);
          child.snapshot.input.stepNode.config = {
            ...child.snapshot.input.stepNode.config,
            ...stepDef
          };
          child.snapshot.input.context = parsedSnapshot.context;
        }
      });
    }
    const updateStepInHierarchy = (value, targetStepId) => {
      const result = {};
      for (const key of Object.keys(value)) {
        const currentValue = value[key];
        if (key === targetStepId) {
          result[key] = "pending";
        } else if (typeof currentValue === "object" && currentValue !== null) {
          result[key] = updateStepInHierarchy(currentValue, targetStepId);
        } else {
          result[key] = currentValue;
        }
      }
      return result;
    };
    parsedSnapshot.value = updateStepInHierarchy(parsedSnapshot.value, stepId);
    if (parsedSnapshot.context?.attempts) {
      parsedSnapshot.context.attempts[stepId] = this.#steps[stepId]?.retryConfig?.attempts || this.#retryConfig?.attempts || 3;
    }
    this.logger.debug("Resuming workflow with updated snapshot", {
      updatedSnapshot: parsedSnapshot,
      runId,
      stepId
    });
    return this.execute({
      snapshot: parsedSnapshot,
      runId,
      stepId
    });
  }
  __registerPrimitives(p) {
    if (p.telemetry) {
      this.__setTelemetry(p.telemetry);
    }
    if (p.logger) {
      this.__setLogger(p.logger);
    }
    this.#mastra = p;
  }
  get stepGraph() {
    return this.#stepGraph;
  }
  get stepSubscriberGraph() {
    return this.#stepSubscriberGraph;
  }
  get steps() {
    return this.#steps;
  }
};

export { Step, Workflow, createStep, getStepResult, isErrorEvent, isTransitionEvent, isVariableReference };
